//***********************************************************
//
//  NXV4 "Math-on-Paper" software
//
//  package : MTX class _____________________________________
//            
//            supports: move semantix
//                      operators + - * / L,R-val context
//                                += -= *= /=
//                      fabs()
//                      ~  -      unary operators L,R-val
//                      mtx<arithmetic> i.e.     int
//                                               float 
//                                               double
//                                               real
//                                           cpx<int> 
//                                           cpx<float> 
//                                           cpx<double> 
//                                           cpx<real> 
//                      type_casts between the types 
//                      cout << overload 
//
//                      TOTAL = 5084 instantiated op/f's
//                      _____________________________________
//
//
//  authors : Mihai-Tiberiu Dima (1)
//            Maria Dima         (2)
//
//             Meshcheryakov Laboratory of Information Techn.
//             Joint Institute for Nuclear Research
//             Dubna-Russia
//
//            on leave from:
//
//             Hyperion Univ. Bucharest-Romania
//
//
//  date    : MTX4/v1.0.alfa   / Mon Mar 07 12:45:54 CET 2022
//
//***********************************************************
#pragma once


#include<iostream>
#include<cmath>
#include<type_traits>
#include<utility>

#include "cpx.hh"
#include "vec.hh"

using namespace cpx4                                        ;
using namespace vec4                                        ;

using std::cout                                             ;
using std::endl                                             ;
using std::decay                                            ;
using std::declval                                          ;
using std::enable_if                                        ;
using std::is_same                                          ;
using std::forward                                          ;
using std::ostream                                          ;
using std::integral_constant                                ;
using std::remove_cv                                        ;


namespace mtx4
{

 bool  utf8x = false                                        ; 

 using fltx =      float                                    ;
 using dblx =      double                                   ;
 using real = long double                                   ;

template< class T >
struct is_cscalar: integral_constant<bool, 
 is_same<    int         , typename remove_cv<T>::type>::value ||
 is_same<    fltx        , typename remove_cv<T>::type>::value ||
 is_same<    dblx        , typename remove_cv<T>::type>::value ||
 is_same<    real        , typename remove_cv<T>::type>::value ||
 is_same<cpx<int >       , typename remove_cv<T>::type>::value ||
 is_same<cpx<fltx>       , typename remove_cv<T>::type>::value ||
 is_same<cpx<dblx>       , typename remove_cv<T>::type>::value ||
 is_same<cpx<real>       , typename remove_cv<T>::type>::value  >
{                                                           ;};


 template<typename scalar=dblx>
 class mtx
   {public: typename decay<scalar>::type x11 = 0            ;
            typename decay<scalar>::type x12 = 0            ;
            typename decay<scalar>::type x13 = 0            ;
            typename decay<scalar>::type x21 = 0            ;
            typename decay<scalar>::type x22 = 0            ;
            typename decay<scalar>::type x23 = 0            ;
            typename decay<scalar>::type x31 = 0            ;
            typename decay<scalar>::type x32 = 0            ;
            typename decay<scalar>::type x33 = 0            ;
      // 
      mtx()                         // plain     
       {                                                    ;} 
      //
      mtx(scalar const& a)          // manual     
       {x11 = a                                             ;
        x12 = 0                                             ;
        x13 = 0                                             ;
        x21 = 0                                             ;
        x22 = a                                             ;
        x23 = 0                                             ;
        x31 = 0                                             ;
        x32 = 0                                             ;
        x33 = a                                             ;}
      //
      template<typename U>
      mtx(U const& a)
       {x11 = scalar(const_cast<U&>(a))                     ;
        x12 = 0                                             ;
        x13 = 0                                             ;
        x21 = 0                                             ;
        x22 = scalar(const_cast<U&>(a))                     ;
        x23 = 0                                             ;
        x31 = 0                                             ;
        x32 = 0                                             ;
        x33 = scalar(const_cast<U&>(a))                     ;
       }
      //
      mtx(scalar const& a11, scalar const& a12, scalar const& a13,
          scalar const& a21, scalar const& a22, scalar const& a23,
          scalar const& a31, scalar const& a32, scalar const& a33 )
       {x11 = a11                                           ;
        x12 = a12                                           ;
        x13 = a13                                           ;
        x21 = a21                                           ;
        x22 = a22                                           ;
        x23 = a23                                           ;
        x31 = a31                                           ;
        x32 = a32                                           ;
        x33 = a33                                           ;}
      //
      template<typename U>
      mtx(U const& a11, U const& a12, U const& a13,
          U const& a21, U const& a22, U const& a23,
          U const& a31, U const& a32, U const& a33 )
       {x11 = scalar(const_cast<U&>(a11))                   ;
        x12 = scalar(const_cast<U&>(a12))                   ;
        x13 = scalar(const_cast<U&>(a13))                   ;
        x21 = scalar(const_cast<U&>(a21))                   ;
        x22 = scalar(const_cast<U&>(a22))                   ;
        x23 = scalar(const_cast<U&>(a23))                   ;
        x31 = scalar(const_cast<U&>(a31))                   ;
        x32 = scalar(const_cast<U&>(a32))                   ;
        x33 = scalar(const_cast<U&>(a33))                   ;
       }
      //
      mtx(const mtx& A)             // copy       
       {x11 = A.x11                                         ;
        x12 = A.x12                                         ;
        x13 = A.x13                                         ;
        x21 = A.x21                                         ;
        x22 = A.x22                                         ;
        x23 = A.x23                                         ;
        x31 = A.x31                                         ;
        x32 = A.x32                                         ;
        x33 = A.x33                                         ;}

      template<typename U>
      mtx(mtx<U> const& A)
       {x11 = scalar(const_cast<U&>(A.x11))                 ;
        x12 = scalar(const_cast<U&>(A.x12))                 ;
        x13 = scalar(const_cast<U&>(A.x13))                 ;
        x21 = scalar(const_cast<U&>(A.x21))                 ;
        x22 = scalar(const_cast<U&>(A.x22))                 ;
        x23 = scalar(const_cast<U&>(A.x23))                 ;
        x31 = scalar(const_cast<U&>(A.x31))                 ;
        x32 = scalar(const_cast<U&>(A.x32))                 ;
        x33 = scalar(const_cast<U&>(A.x33))                 ;
       }

      // CAST  _____________________________________________

      explicit
      operator     int  ()         
      {return      int  (int((x11+x22+x33)/3))              ;}

      explicit
      operator     fltx ()         
      {return      fltx (fltx((x11+x22+x33)/3))             ;}

      explicit
      operator     dblx ()         
      {return      fltx (dblx((x11+x22+x33)/3))             ;}

      explicit
      operator     real ()         
      {return      fltx (real((x11+x22+x33)/3))             ;}

      explicit
      operator cpx<int >()         
      {return  cpx<int >(int((x11+x22+x33)/3))              ;}

      explicit
      operator cpx<fltx>()         
      {return  cpx<fltx>(fltx((x11+x22+x33)/3))             ;}

      explicit
      operator cpx<dblx>()         
      {return  cpx<dblx>(dblx((x11+x22+x33)/3))             ;}

      explicit
      operator cpx<real>()         
      {return  cpx<real>(real((x11+x22+x33)/3))             ;}

      explicit
      operator mtx<int >()
      {return  mtx<int >(int (x11), int (x12), int (x13),
                         int (x21), int (x22), int (x23),
                         int (x31), int (x32), int (x33) )  ;}
      explicit
      operator mtx<fltx>()
      {return  mtx<fltx>(fltx(x11), fltx(x12), fltx(x13),
                         fltx(x21), fltx(x22), fltx(x23),
                         fltx(x31), fltx(x32), fltx(x33) )  ;}
      explicit
      operator mtx<dblx>()
      {return  mtx<dblx>(dblx(x11), dblx(x12), dblx(x13),
                         dblx(x21), dblx(x22), dblx(x23),
                         dblx(x31), dblx(x32), dblx(x33) )  ;}
      explicit
      operator mtx<real>()
      {return  mtx<real>(real(x11), real(x12), real(x13),
                         real(x21), real(x22), real(x23),
                         real(x31), real(x32), real(x33) )  ;}

      explicit
      operator mtx<cpx<int >>()
      {return  mtx<cpx<int >>(cpx<int >(x11), cpx<int >(x12), cpx<int >(x13),
                              cpx<int >(x21), cpx<int >(x22), cpx<int >(x23),
                              cpx<int >(x31), cpx<int >(x32), cpx<int >(x33) );}
      explicit
      operator mtx<cpx<fltx>>()
      {return  mtx<cpx<fltx>>(cpx<fltx>(x11), cpx<fltx>(x12), cpx<fltx>(x13),
                              cpx<fltx>(x21), cpx<fltx>(x22), cpx<fltx>(x23),
                              cpx<fltx>(x31), cpx<fltx>(x32), cpx<fltx>(x33) );}
      explicit
      operator mtx<cpx<dblx>>()
      {return  mtx<cpx<dblx>>(cpx<dblx>(x11), cpx<dblx>(x12), cpx<dblx>(x13),
                              cpx<dblx>(x21), cpx<dblx>(x22), cpx<dblx>(x23),
                              cpx<dblx>(x31), cpx<dblx>(x32), cpx<dblx>(x33) );}
      explicit
      operator mtx<cpx<real>>()
      {return  mtx<cpx<real>>(cpx<real>(x11), cpx<real>(x12), cpx<real>(x13),
                              cpx<real>(x21), cpx<real>(x22), cpx<real>(x23),
                              cpx<real>(x31), cpx<real>(x32), cpx<real>(x33) );}

      // ASSGN _____________________________________________

      scalar& operator= (const scalar& A)
       {x11 = A                                             ;
        x12 = 0                                             ;
        x13 = 0                                             ;
        x21 = 0                                             ;
        x22 = A                                             ;
        x23 = 0                                             ;
        x31 = 0                                             ;
        x32 = 0                                             ;
        x33 = A                                             ;
        return const_cast<scalar&>(A)                       ;}

         mtx& operator= (const    mtx& A)
       {x11 = A.x11                                         ;
        x12 = A.x12                                         ;
        x13 = A.x13                                         ;
        x21 = A.x21                                         ;
        x22 = A.x22                                         ;
        x23 = A.x23                                         ;
        x31 = A.x31                                         ;
        x32 = A.x32                                         ;
        x33 = A.x33                                         ;
        return const_cast<mtx&>(A)                          ;}

      template<typename U> 
      mtx<U>& operator= (mtx<U> const& A)
       {x11 = scalar(const_cast<U&>(A.x11))                 ;
        x12 = scalar(const_cast<U&>(A.x12))                 ;
        x13 = scalar(const_cast<U&>(A.x13))                 ;
        x21 = scalar(const_cast<U&>(A.x21))                 ;
        x22 = scalar(const_cast<U&>(A.x22))                 ;
        x23 = scalar(const_cast<U&>(A.x23))                 ;
        x31 = scalar(const_cast<U&>(A.x31))                 ;
        x32 = scalar(const_cast<U&>(A.x32))                 ;
        x33 = scalar(const_cast<U&>(A.x33))                 ;
        return const_cast<mtx<U>&>(A)                       ;}

      // NGTV ______________________________________________

      template<typename T>
      friend
      mtx<T>
      operator- (mtx<T> const& A)                           ;

      template<typename T>
      friend
      mtx<T>&&
      operator- (mtx<T> && A)                               ;

      // TILDE (mtx.conj.) _________________________________

      template<typename T>
      friend
      mtx<T>
      operator~ (mtx<T> const& A)                           ;

      template<typename T>
      friend
      mtx<T>&&
      operator~ (mtx<T> && A)                               ;

      // DET / TR __________________________________________

      template<typename T>
      friend
      T tr(mtx<T> const& A)                                 ;

      template<typename T>
      friend
      T det(mtx<T> const& A)                                ;

      // FABS ______________________________________________

      template<typename T>
      friend
      long double fabs(mtx<T> const& A)                     ;

      template<typename T>
      friend
      long double n2(mtx<T> const& A)                       ;

      // PROD ______________________________________________

      template<typename L, typename R>
      friend
      typename decay<decltype(declval<L>()*
                              declval<R>())>::type
      operator|(mtx<L> const& A, mtx<R> const& B)           ;

      template<typename ALX, typename BLX>
      friend
      vec<typename decay<decltype(declval<ALX>()+
                                  declval<BLX>() )
                         >::type
         >
      operator|(vec<ALX> const& a, mtx<BLX> const& b)       ;

      template<typename ALX, typename BLX>
      friend
      vec<typename decay<decltype(declval<ALX>()+
                                  declval<BLX>() )
                         >::type
         >
      operator|(mtx<ALX> const& a, vec<BLX> const& b)       ;

      // OPERATORS __________________________________________

         #include "eqal.cxx"
         #include "comp.cxx"

         #include "plus.cxx"
         #include "mnus.cxx"
         #include "mult.cxx"
         #include "divd.cxx"

         #include "alog.cxx"

      // PRINT _____________________________________________

      friend std::ostream& operator<<(std::ostream&   , 
                                      mtx<auto> const& )    ;

      // ___________________________________________________

     ~mtx()
       {                                                    ;}
 };

//===========================================================
};
