//***********************************************************
//
//  NXV4 "Math-on-Paper" software
//
//  package : VEC class _____________________________________
//            
//            supports: move semantix
//                      operators + - * L,R-val context
//                                += -= *= /=
//                      fabs()
//                      ~  -      unary operators L,R-val
//                      mtx<arithmetic> i.e.     int
//                                               float 
//                                               double
//                                               real
//                                           cpx<int> 
//                                           cpx<float> 
//                                           cpx<double> 
//                                           cpx<real> 
//                      type_casts between the types 
//                      cout << overload 
//
//                      TOTAL = 2836 instantiated op/f's
//                      _____________________________________
//
//
//  authors : Mihai-Tiberiu Dima (1)
//            Maria Dima         (2)
//
//             Meshcheryakov Laboratory of Information Techn.
//             Joint Institute for Nuclear Research
//             Dubna-Russia
//
//            on leave from:
//
//             Hyperion Univ. Bucharest-Romania
//
//
//  date    : VEC4/v1.0.alfa   / Mon Mar 07 12:45:54 CET 2022
//
//***********************************************************
#include <cmath>
#include <iostream>

#include "cpx.hh"
#include "vec.hh"

using namespace std                                         ;
using namespace cpx4                                        ;
using namespace vec4                                        ;


int main()
{

 vec4::utf8x = true                                         ;

// this is a logical operator example
//
//   |chi> = -|up> + 5*i*|down>
//   |vec> = 3|up> + 2*i*|down>
//
//   || chi ||  >  || vec ||
//
// to use this example do the following:
//   rm test.cc
//   ln -s e1.cc test.cc
//   make test
//   make run

  using fltx = float                                        ;
  using dblx = double                                       ;
  using real = long double                                  ;

  double e   = exp(1)                                       ;
  double pi  = 3.14159265853                                ;

  cpx<int >  i(0,1)                                         ;

  auto chi = vec<int>(-1,0,1) + 5*i*vec<dblx>(1,0,1)        ; 
  auto psi = vec<int>( 3,0,2) + 2*i*vec<dblx>(3,0,1)        ; 

  cout                                             << endl  ;
  cout << noboolalpha                              << endl  ;

  cout                                             << endl  ;
  cout << " _____________________________________" << endl  ;
  cout                                             << endl  ;
  cout << "   " << noboolalpha
                << "is  || "      << chi
                << " ||  >  || "  << psi
                << " ||  ? "      << endl << endl
                << "   logical value: "      
                << (chi > psi)                     << endl  ;
  cout << " _____________________________________" << endl  ;
  cout                                             << endl  ;
  cout                                             << endl  ;
  cout                                             << endl  ;

}
